import betfair_data
import logging

logging.basicConfig(level=logging.WARN, format='%(levelname)s %(name)s %(message)s')

paths = [
    "data/2021_10_OctRacingAUPro.tar",
    "data/2021_11_NovRacingAUPro.tar",
    "data/2021_12_DecRacingAUPro.tar",
]

market_count = 0
update_count = 0

for market in betfair_data.TarBz2(paths).mutable():
    market_count += 1
    update_count += 1 # market has inital update already done
    
    while market.update():
        update_count += 1

    print(f"Markets {market_count} Updates {update_count}", end='\r')
print(f"Markets {market_count} Updates {update_count}")



from typing import Sequence 

import unittest.mock
import tarfile
import bz2
import betfairlightweight


trading = betfairlightweight.APIClient("username", "password", "appkey")
listener = betfairlightweight.StreamListener(
    max_latency=None, lightweight=False, update_clk=False, output_queue=None, cumulative_runner_tv=True, calculate_market_tv=True
)

paths = [ 
    "data/2021_10_OctRacingAUPro.tar",
    "data/2021_11_NovRacingAUPro.tar",
    "data/2021_12_DecRacingAUPro.tar"
]

def load_tar(file_paths: Sequence[str]):
    for file_path in file_paths:
        with tarfile.TarFile(file_path) as archive:
            for file in archive:
                yield bz2.open(archive.extractfile(file)).read()
    return None

market_count = 0
update_count = 0

for i, file_obj in enumerate(load_tar(paths)):
    with unittest.mock.patch("builtins.open", lambda f, _: f):  
        stream = trading.streaming.create_historical_generator_stream(
            file_path=file_obj,
            listener=listener,
        )
        gen = stream.get_generator()

        market_count += 1
        for market_books in gen():
            for market_book in market_books:
                update_count += 1

    print(f"Market {market_count} Update {update_count}", end='\r')
print(f"Market {market_count} Update {update_count}")



import tarfile
import bz2
from typing import Sequence
import betfair_data
import logging
from betfair_data import bflw

logging.basicConfig(level=logging.WARN, format='%(levelname)s %(name)s %(message)s')

paths = [ 
    "data/2021_10_OctRacingAUPro.tar",
    "data/2021_11_NovRacingAUPro.tar",
    "data/2021_12_DecRacingAUPro.tar"
]

def run_with_py_loading():
    market_count = 0
    update_count = 0

    def load_tar(file_paths: Sequence[str]):
        for file_path in file_paths:
            with tarfile.TarFile(file_path) as archive:
                for file in archive:
                    f: bz2.BZ2File = bz2.open(archive.extractfile(file))
                    name = file.name
                    bytes = f.read()

                    yield (name, bytes)
        return None

    for name, bytes in load_tar(paths):
        file = bflw.BflwIter(name, bytes, cumulative_runner_tv=True)

        market_count += 1
        for market_books in file:
            for market_book in market_books:
                update_count += 1

        print(f"Market {market_count} Update {update_count}", end='\r')
    print(f"Market {market_count} Update {update_count}")



def run_with_rust_loading():
    market_count = 0
    update_count = 0
    
    for file in betfair_data.TarBz2(paths, cumulative_runner_tv=True).bflw():
        market_count += 1

        for market_books in file:
            for market_book in market_books:
                update_count += 1

        print(f"Market {market_count} Update {update_count}", end='\r')
    print(f"Market {market_count} Update {update_count}")


run_with_rust_loading()


from typing import Iterator, List, Sequence, Optional
from betfair_data import bflw

class MarketImage():
    """
    A class representing a Betfair Market.
    """
    source: str
    file: str
    market_id: str
    """Market Id - the id of the market"""
    bet_delay: int
    betting_type: str
    bsp_market: bool
    """Is BSP betting available for this market
            >>> m.bsp_market
            True
    """
    bsp_reconciled: bool
    """Has the starting price been detirmined for this market
            >>> m.bsp_reconciled
            False
    """
    clk: str
    """Token value (non-null) should be stored and passed in a MarketSubscriptionMessage to resume subscription (in case of disconnect)"""
    complete: bool
    """Is the market in a completed state
            >>> m.complete
            False
    """
    country_code: str
    """The country the market is taking place in. 2 digit string
            >>> m.country_code
            'AU'
    """
    cross_matching: bool
    discount_allowed: bool
    each_way_divisor: Optional[float]
    event_id: int
    event_name: Optional[str]
    event_type_id: int
    in_play: bool
    market_base_rate: int
    market_name: Optional[str]
    market_time: int
    market_type: str
    number_of_active_runners: int
    number_of_winners: int
    open_date: int
    persistence_enabled: bool
    publish_time: int
    """Publish Time (in millis since epoch) that the changes were generated"""
    runners_voidable: bool
    runners: List[Runner]   
    settled_time: Optional[int]
    status: str
    """The markets current status, value is one of ["INACTIVE", "OPEN", "SUSPENDED", "CLOSED"]
        >>> market.status
        'CLOSED'
    """
    suspend_time: Optional[int]
    timezone: str
    total_matched: float
    """The total amount matched across the market. This value is truncated at 2dp (or null if un-changed)
        >>> m.total_matched
        53212.45
    """
    turn_in_play_enabled: bool
    venue: Optional[str]
    version: int

class Market(MarketImage):
    """"""
    def __init__(self, path: str, bytes: bytes, cumulative_runner_tv: bool = True, stable_runner_index = True) -> None: ...

    def update(self) -> bool: 
        """ Update the market with the next delta

        Example:
            >>> market.update()
        """  
    def copy(self) -> MarketImage:
        """ Create an immutable copy of the market, including runners and prices that wont be updated

        Example:
            >>> c = market.copy()
            >>> market.update()
            >>> market.publish_time != c.publish_time
            True
        """  

class Runner():
    """
    A class representing a Betfair Runner.
    """
    selection_id: int
    name: Optional[str]
    status: str
    """The runners current status, value is one of ["ACTIVE", "REMOVED", "REMOVED_VACANT", "WINNER", "PLACED", "LOSER", "HIDDEN"]
        >>> market.runners[0].status
        'ACTIVE'
    """
    last_price_traded: Optional[float]
    total_matched: float
    adjustment_factor: Optional[float]
    handicap: Optional[float]
    sort_priority: int
    removal_date: Optional[int]
    ex: RunnerBookEX
    sp: RunnerBookSP

class RunnerBookEX():
    available_to_back: List[PriceSize]
    available_to_lay: List[PriceSize]
    traded_volume: List[PriceSize]

class RunnerBookSP(): 
    far_price: Optional[float]
    near_price: Optional[float]
    actual_sp: Optional[float]
    back_stake_taken: List[PriceSize]
    lay_liability_taken: List[PriceSize]

class PriceSize():
    price: float
    size: float

# sources
class MutableAdapter(Iterator[Market]): ...

class TarBz2():
    """"""
    def __init__(self, paths: Sequence[str], cumulative_runner_tv: bool = True) -> None: ...

    def mutable(self, stable_runner_index = True) -> MutableAdapter: ...
    def bflw(self) -> bflw.BflwAdapter: ...


class Files():
    """"""
    def __init__(self, paths: Sequence[str], cumulative_runner_tv: bool = True) -> None: ...

    def mutable(self, stable_runner_index = True) -> MutableAdapter: ...
    def bflw(self) -> bflw.BflwAdapter: ...



from glob import glob
from types import NoneType
from typing import List, Sequence 
import tarfile
import bz2
import math
import gzip
import os
import unittest.mock
import betfairlightweight
import betfair_data

# types
from betfair_data import bflw as bfd
import betfairlightweight.resources.bettingresources as bflw
import betfairlightweight.resources.streamingresources as bflws

#  paths to test files
paths = glob("data/stream/*.gz")
# paths = [
#     "data/2021_12_DecRacingAUPro.tar",
#     "data/2021_10_OctRacingAUPro.tar",
#     "data/2021_11_NovRacingAUPro.tar",
# ]
# find way to garentee order - this is gross
bfd_source = betfair_data.Files(paths, cumulative_runner_tv=True).bflw()

def bflw_source(file_paths: Sequence[str]):
    trading = betfairlightweight.APIClient("username", "password", "appkey")
    listener = betfairlightweight.StreamListener(
        max_latency=None, update_clk=False, output_queue=None, cumulative_runner_tv=True, calculate_market_tv=True
    )

    for file_path in file_paths:
        if os.path.isfile(file_path):
            ext = os.path.splitext(file_path)[1]
            if ext == '.tar':
                with tarfile.TarFile(file_path) as archive:
                    for file in archive:
                        f = bz2.open(archive.extractfile(file))
                        with unittest.mock.patch("builtins.open", lambda f, _: f):  
                            stream = trading.streaming.create_historical_generator_stream(
                                file_path=f,
                                listener=listener,
                            )
                            yield stream.get_generator()
            elif ext == '.gz':
                with gzip.open(file_path, mode='r') as f:   
                    with unittest.mock.patch("builtins.open", lambda f, _: f):  
                        stream = trading.streaming.create_historical_generator_stream(
                            file_path=f,
                            listener=listener,
                        )
                        yield stream.get_generator()
    return None

# run the tests, panics out on fail
def run_test():
    updates = 0
    files = 0
    for (bfd_iter, bflw_gen) in zip(bfd_source, bflw_source(paths)): 
        files += 1       
        for (bfd_mbs, bflw_mbs) in zip(bfd_iter, bflw_gen()):
            for (bfd_m, bflw_m) in zip(bfd_mbs, bflw_mbs):
                bflw_m: bflw.MarketBook
                updates += 1
                print(f"Updates Tested {bfd_m.market_id} files:{files} updates:{updates}", end='\r')
                test_MarketBook(bfd_m, bflw_m)
    print(f"Updates Tested files:{files} updates:{updates}")

def test_MarketBook(bfd_m: bfd.MarketBook, bflw_m: bflw.MarketBook):
    assert bfd_m.bet_delay == bflw_m.bet_delay, f"<MarketBook> bet_delay {bfd_m.bet_delay} != {bflw_m.bet_delay}"
    assert bfd_m.bsp_reconciled == bflw_m.bsp_reconciled, f"<MarketBook> bsp_reconciled {bfd_m.bsp_reconciled} != {bflw_m.bsp_reconciled}"
    assert bfd_m.complete == bflw_m.complete, f"<MarketBook> complete {bfd_m.complete} != {bflw_m.complete}"
    assert bfd_m.cross_matching == bflw_m.cross_matching, f"<MarketBook> cross_matching {bfd_m.cross_matching} != {bflw_m.cross_matching}"
    assert bfd_m.inplay == bflw_m.inplay, f"<MarketBook> inplay {bfd_m.inplay} != {bflw_m.inplay}"
    assert bfd_m.is_market_data_delayed == bflw_m.is_market_data_delayed, f"<MarketBook> is_market_data_delayed {bfd_m.is_market_data_delayed} != {bflw_m.is_market_data_delayed}"
    assert bfd_m.last_match_time == bflw_m.last_match_time, f"<MarketBook> last_match_time {bfd_m.last_match_time} != {bflw_m.last_match_time}"
    assert bfd_m.market_id == bflw_m.market_id, f"<MarketBook> market_id {bfd_m.market_id} != {bflw_m.market_id}"
    assert bfd_m.number_of_active_runners == bflw_m.number_of_active_runners, f"<MarketBook> number_of_active_runners {bfd_m.number_of_active_runners} != {bflw_m.number_of_active_runners}"
    assert bfd_m.number_of_runners == bflw_m.number_of_runners, f"<MarketBook> number_of_runners {bfd_m.number_of_runners} != {bflw_m.number_of_runners}"
    assert bfd_m.number_of_winners == bflw_m.number_of_winners, f"<MarketBook> number_of_winners {bfd_m.number_of_winners} != {bflw_m.number_of_winners}"
    assert bfd_m.publish_time == bflw_m.publish_time, f"<MarketBook> publish_time {bfd_m.publish_time} != {bflw_m.publish_time}"
    assert bfd_m.publish_time_epoch == bflw_m.publish_time_epoch, f"<MarketBook> publish_time_epoch {bfd_m.publish_time_epoch} != {bflw_m.publish_time_epoch}"
    assert bfd_m.runners_voidable == bflw_m.runners_voidable, f"<MarketBook> runners_voidable {bfd_m.runners_voidable} != {bflw_m.runners_voidable}"
    assert bfd_m.status == bflw_m.status, f"<MarketBook> status {bfd_m.status} != {bflw_m.status}"
    assert bfd_m.total_available == bflw_m.total_available, f"<MarketBook> total_available {bfd_m.total_available} != {bflw_m.total_available}"
    assert bfd_m.total_matched == bflw_m.total_matched, f"<MarketBook> total_matched {bfd_m.total_matched} != {bflw_m.total_matched}"
    assert bfd_m.version == bflw_m.version, f"<MarketBook> version {bfd_m.version} != {bflw_m.version}"
    test_MarketDefinition(bfd_m.market_definition, bflw_m.market_definition)
    for bfd_rb, bflw_rb in zip(bfd_m.runners, bflw_m.runners):
        test_RunnerBook(bfd_rb, bflw_rb) 

def test_MarketDefinition(bfd_md: bfd.MarketDefinition, bflw_md: bflws.MarketDefinition):
    assert bfd_md.bet_delay == bflw_md.bet_delay, f"<MarketDefinition> bet_delay {bfd_md.bet_delay} != {bflw_md.bet_delay}"
    assert bfd_md.betting_type == bflw_md.betting_type, f"<MarketDefinition> betting_type {bfd_md.betting_type} != {bflw_md.betting_type}"
    assert bfd_md.bsp_market == bflw_md.bsp_market, f"<MarketDefinition> bsp_market {bfd_md.bsp_market} != {bflw_md.bsp_market}"
    assert bfd_md.bsp_reconciled == bflw_md.bsp_reconciled, f"<MarketDefinition> bsp_reconciled {bfd_md.bsp_reconciled} != {bflw_md.bsp_reconciled}"
    assert bfd_md.complete == bflw_md.complete, f"<MarketDefinition> complete {bfd_md.complete} != {bflw_md.complete}"
    assert bfd_md.country_code == bflw_md.country_code, f"<MarketDefinition> country_code {bfd_md.country_code} != {bflw_md.country_code}"
    assert bfd_md.cross_matching == bflw_md.cross_matching, f"<MarketDefinition> cross_matching {bfd_md.cross_matching} != {bflw_md.cross_matching}"
    assert bfd_md.discount_allowed == bflw_md.discount_allowed, f"<MarketDefinition> discount_allowed {bfd_md.discount_allowed} != {bflw_md.discount_allowed}"
    assert bfd_md.event_id == bflw_md.event_id, f"<MarketDefinition> event_id {bfd_md.event_id} != {bflw_md.event_id}"
    assert bfd_md.event_type_id == bflw_md.event_type_id, f"<MarketDefinition> event_type_id {bfd_md.event_type_id} != {bflw_md.event_type_id}"
    assert bfd_md.in_play == bflw_md.in_play, f"<MarketDefinition> in_play {bfd_md.in_play} != {bflw_md.in_play}"
    assert bfd_md.market_base_rate == bflw_md.market_base_rate, f"<MarketDefinition> market_base_rate {bfd_md.market_base_rate} != {bflw_md.market_base_rate}"
    assert bfd_md.market_time == bflw_md.market_time, f"<MarketDefinition> market_time {bfd_md.market_time} != {bflw_md.market_time}"
    assert bfd_md.market_type == bflw_md.market_type, f"<MarketDefinition> market_type {bfd_md.market_type} != {bflw_md.market_type}"
    assert bfd_md.number_of_active_runners == bflw_md.number_of_active_runners, f"<MarketDefinition> number_of_active_runners {bfd_md.number_of_active_runners} != {bflw_md.number_of_active_runners}"
    assert bfd_md.number_of_winners == bflw_md.number_of_winners, f"<MarketDefinition> number_of_winners {bfd_md.number_of_winners} != {bflw_md.number_of_winners}"
    assert bfd_md.open_date == bflw_md.open_date, f"<MarketDefinition> open_date {bfd_md.open_date} != {bflw_md.open_date}"
    assert bfd_md.persistence_enabled == bflw_md.persistence_enabled, f"<MarketDefinition> persistence_enabled {bfd_md.persistence_enabled} != {bflw_md.persistence_enabled}"
    assert bfd_md.regulators == bflw_md.regulators, f"<MarketDefinition> regulators {bfd_md.regulators} != {bflw_md.regulators}"
    assert bfd_md.runners_voidable == bflw_md.runners_voidable, f"<MarketDefinition> runners_voidable {bfd_md.runners_voidable} != {bflw_md.runners_voidable}"
    assert bfd_md.settled_time == bflw_md.settled_time, f"<MarketDefinition> settled_time {bfd_md.settled_time} != {bflw_md.settled_time}"
    assert bfd_md.status == bflw_md.status, f"<MarketDefinition> status {bfd_md.status} != {bflw_md.status}"
    assert bfd_md.suspend_time == bflw_md.suspend_time, f"<MarketDefinition> suspend_time {bfd_md.suspend_time} != {bflw_md.suspend_time}"
    assert bfd_md.timezone == bflw_md.timezone, f"<MarketDefinition> timezone {bfd_md.timezone} != {bflw_md.timezone}"
    assert bfd_md.turn_in_play_enabled == bflw_md.turn_in_play_enabled, f"<MarketDefinition> turn_in_play_enabled {bfd_md.turn_in_play_enabled} != {bflw_md.turn_in_play_enabled}"
    assert bfd_md.venue == bflw_md.venue, f"<MarketDefinition> venue {bfd_md.venue} != {bflw_md.venue}"
    assert bfd_md.version == bflw_md.version, f"<MarketDefinition> version {bfd_md.version} != {bflw_md.version}"
    for bfd_rd, bflw_rd in zip(bfd_md.runners, bflw_md.runners):
        test_MarketDefinitionRunner(bfd_rd, bflw_rd) 

def test_MarketDefinitionRunner(bfd_rd: bfd.MarketDefinitionRunner, bflw_rd: bflws.MarketDefinitionRunner):
    assert bfd_rd.selection_id == bflw_rd.selection_id, f"<MarketDefinitionRunner> selection_id {bfd_rd.selection_id} != {bflw_rd.selection_id}"
    assert test_float(bfd_rd.adjustment_factor, bflw_rd.adjustment_factor), f"<MarketDefinitionRunner> adjustment_factor {bfd_rd.adjustment_factor} != {bflw_rd.adjustment_factor}"
    assert bfd_rd.removal_date == bflw_rd.removal_date, f"<MarketDefinitionRunner> removal_date {bfd_rd.removal_date} != {bflw_rd.removal_date}"
    assert bfd_rd.sort_priority == bflw_rd.sort_priority, f"<MarketDefinitionRunner> sort_priority {bfd_rd.sort_priority} != {bflw_rd.sort_priority}"
    assert bfd_rd.status == bflw_rd.status, f"<MarketDefinitionRunner> status {bfd_rd.status} != {bflw_rd.status}"
    assert test_float(bfd_rd.bsp, bflw_rd.bsp), f"<MarketDefinitionRunner> bsp {bfd_rd.bsp} != {bflw_rd.bsp}"
    assert test_float(bfd_rd.handicap, bflw_rd.handicap), f"<MarketDefinitionRunner> handicap {bfd_rd.handicap} != {bflw_rd.handicap}"
    assert bfd_rd.name == bflw_rd.name, f"<MarketDefinitionRunner> name {bfd_rd.name} != {bflw_rd.name}"
    
def test_RunnerBook(bfd_rb: bfd.RunnerBook, bflw_rb: bflw.RunnerBook):
    assert bfd_rb.selection_id == bflw_rb.selection_id, f"<RunnerBook> selection_id {bfd_rb.selection_id} != {bflw_rb.selection_id}"
    assert bfd_rb.adjustment_factor == bflw_rb.adjustment_factor, f"<RunnerBook> adjustment_factor {bfd_rb.adjustment_factor} != {bflw_rb.adjustment_factor}"
    assert bfd_rb.handicap == bflw_rb.handicap, f"<RunnerBook> handicap {bfd_rb.handicap} != {bflw_rb.handicap}"
    assert bfd_rb.last_price_traded == bflw_rb.last_price_traded, f"<RunnerBook> last_price_traded {bfd_rb.last_price_traded} != {bflw_rb.last_price_traded}"
    assert bfd_rb.removal_date == bflw_rb.removal_date, f"<RunnerBook> removal_date {bfd_rb.removal_date} != {bflw_rb.removal_date}"
    assert bfd_rb.status == bflw_rb.status, f"<RunnerBook> status {bfd_rb.status} != {bflw_rb.status}"
    assert bfd_rb.total_matched == bflw_rb.total_matched, f"<RunnerBook> total_matched {bfd_rb.total_matched} != {bflw_rb.total_matched}"
    assert bfd_rb.matches == bflw_rb.matches, f"<RunnerBook> matches {bfd_rb.matches} != {bflw_rb.matches}"
    assert bfd_rb.orders == bflw_rb.orders, f"<RunnerBook> orders {bfd_rb.orders} != {bflw_rb.orders}"
    test_RunnerBookEX(bfd_rb.ex, bflw_rb.ex, bfd_rb.selection_id)
    test_RunnerBookSP(bfd_rb.sp, bflw_rb.sp, bfd_rb.selection_id)
    
def test_RunnerBookEX(bfd_ex: betfair_data.RunnerBookEX, bflw_ex: bflw.RunnerBookEX, sid: int):
    test_ListPriceSize(bfd_ex.available_to_back, bflw_ex.available_to_back, f"<RunnerBookEX[{sid}].available_to_back>")
    test_ListPriceSize(bfd_ex.available_to_lay, bflw_ex.available_to_lay, f"<RunnerBookEX[{sid}].available_to_lay>")
    test_ListPriceSize(bfd_ex.traded_volume, bflw_ex.traded_volume, f"<RunnerBookEX[{sid}].traded_volume>")

def test_RunnerBookSP(bfd_sp: betfair_data.RunnerBookSP, bflw_sp: bflw.RunnerBookSP, sid: int):
    assert test_float(bfd_sp.far_price, bflw_sp.far_price), f"<RunnerBookSP> far_price {bfd_sp.far_price} != {bflw_sp.far_price}"
    assert test_float(bfd_sp.near_price, bflw_sp.near_price), f"<RunnerBookSP> near_price {bfd_sp.near_price} != {bflw_sp.near_price}"
    assert test_float(bfd_sp.actual_sp, bflw_sp.actual_sp), f"<RunnerBookSP> actual_sp {bfd_sp.actual_sp} != {bflw_sp.actual_sp}"
    test_ListPriceSize(bfd_sp.back_stake_taken, bflw_sp.back_stake_taken, f"<RunnerBookSP[{sid}].back_stake_taken>")
    test_ListPriceSize(bfd_sp.lay_liability_taken, bflw_sp.lay_liability_taken, f"<RunnerBookSP[{sid}].lay_liability_taken>")

def test_ListPriceSize(bfd_ps: List[betfair_data.PriceSize], bflw_ps: List[bflw.PriceSize], context: str):
    assert len(bfd_ps) == len(bflw_ps), f"{context} <List[PriceSize]> different lengths {print_ladder(bfd_ps)} != {print_ladder(bflw_ps)}"
    for i, (bfd, bflw) in enumerate(zip(bfd_ps, bflw_ps)):
        assert test_float(bfd.price, bflw.price), f"{context} <List[PriceSize][{i}]> Price  {bfd.price} != {bflw.price}"
        assert test_float(bfd.size, bflw.size), f"{context} <List[PriceSize][{i}]> Size  {bfd.size} != {bflw.size}"
    
def test_float(f1: float|str|NoneType, f2: float|str|int|NoneType) -> bool:
    if isinstance(f2, int):
        f2 = float(f2)
    if isinstance(f1, float):   
        return ((f1 == f2) or math.isclose(f1, f2, abs_tol=0.001) or (math.isnan(f1) and math.isnan(f2)) or (math.isinf(f1) and math.isinf(f2)))

    return f1 == f2

def print_ladder(ladder: List[bflw.PriceSize]) -> str:
    return ' '.join(list(f"[{ps.price}, {ps.size}]" for ps in ladder))


run_test()



import logging
from typing import List, Tuple

from itertools import zip_longest
import functools

from betfair_data import PriceSize
from betfair_data import bflw
import betfair_data

file_output = "output_rust_source.csv"

market_paths = [
    "data/2021_10_OctRacingAUPro.tar",
    "data/2021_11_NovRacingAUPro.tar",
    "data/2021_12_DecRacingAUPro.tar",
]

# setup logging
logging.basicConfig(level=logging.FATAL)

# rounding to 2 decimal places or returning '' if blank
def as_str(v) -> str:
    return '%.2f' % v if (type(v) is float) or (type(v) is int) else v if type(v) is str else ''

# returning smaller of two numbers where min not 0
def min_gr0(a: float, b: float) -> float:
    if a <= 0:
        return b
    if b <= 0:
        return a

    return min(a, b)

# parsing price data and pulling out weighted avg price, matched, min price and max price
def parse_traded(traded: List[PriceSize]) -> Tuple[float, float, float, float]:
    if len(traded) == 0: 
        return (None, None, None, None)

    (wavg_sum, matched, min_price, max_price) = functools.reduce(
        lambda total, ps: (
            total[0] + (ps.price * ps.size), # wavg_sum before we divide by total matched
            total[1] + ps.size, # total matched
            min(total[2], ps.price), # min price matched
            max(total[3], ps.price), # max price matched
        ),
        traded,
        (0, 0, 1001, 0) # starting default values
    )

    wavg_sum = (wavg_sum / matched) if matched > 0 else None # dividing sum of wavg by total matched
    matched = matched if matched > 0 else None 
    min_price = min_price if min_price != 1001 else None
    max_price = max_price if max_price != 0 else None

    return (wavg_sum, matched, min_price, max_price)

# splitting race name and returning the parts 
def split_anz_horse_market_name(market_name: str) -> Tuple[str, str, str]:
    # return race no, length, race type
    # input samples: 
    # 'R6 1400m Grp1' -> ('R6','1400m','grp1')
    # 'R1 1609m Trot M' -> ('R1', '1609m', 'trot')
    # 'R4 1660m Pace M' -> ('R4', '1660m', 'pace')
    parts = market_name.split(' ')
    race_no = parts[0] 
    race_len = parts[1] 
    race_type = parts[2].lower() 

    return (race_no, race_len, race_type)

# filtering markets to those that fit the following criteria
def filter_market(market: bflw.MarketBook) -> bool: 
    d = market.market_definition
    return (d != None
        and d.country_code == 'AU' 
        and d.market_type == 'WIN' 
        and (c := split_anz_horse_market_name(d.name)[2]) != 'trot' and c != 'pace')



# record prices to a file
with open(file_output, "w") as output:
    # defining column headers
    output.write("market_id,event_date,country,track,market_name,selection_id,selection_name,result,bsp,pp_min,pp_max,pp_wap,pp_ltp,pp_volume,ip_min,ip_max,ip_wap,ip_ltp,ip_volume\n")

    for i, g in enumerate(betfair_data.TarBz2(market_paths).bflw()):
        print("Market {}".format(i), end='\r')

        def get_pre_post_final():
            eval_market = None
            prev_market = None
            preplay_market = None
            postplay_market = None       

            for market_books in g:
                for market_book in market_books:
                    # if market doesn't meet filter return out
                    if eval_market is None and ((eval_market := filter_market(market_book)) == False):
                        return (None, None, None)

                    # final market view before market goes in play
                    if prev_market is not None and prev_market.inplay != market_book.inplay:
                        preplay_market = prev_market

                    # final market view at the conclusion of the market
                    if prev_market is not None and prev_market.status == "OPEN" and market_book.status != prev_market.status:
                        postplay_market = market_book

                    # update reference to previous market
                    prev_market = market_book

            return (preplay_market, postplay_market, prev_market) # prev is now final

        (preplay_market, postplay_market, final_market) = get_pre_post_final()

        # no price data for market
        if postplay_market is None:
            continue; 

        preplay_traded = [ (r.last_price_traded, r.ex.traded_volume) for r in preplay_market.runners ] if preplay_market is not None else None
        postplay_traded = [ (
            r.last_price_traded,
            r.ex.traded_volume,
            # calculating SP traded vol as smaller of back_stake_taken or (lay_liability_taken / (BSP - 1))        
            min_gr0(
                next((pv.size for pv in r.sp.back_stake_taken if pv.size > 0), 0),
                next((pv.size for pv in r.sp.lay_liability_taken if pv.size > 0), 0)  / ((r.sp.actual_sp if (type(r.sp.actual_sp) is float) or (type(r.sp.actual_sp) is int) else 0) - 1)
            ) if r.sp.actual_sp is not None else 0,
        ) for r in postplay_market.runners ]

        # generic runner data
        runner_data = [
            {
                'selection_id': r.selection_id,
                'selection_name': next((rd.name for rd in final_market.market_definition.runners if rd.selection_id == r.selection_id), None),
                'selection_status': r.status,
                'sp': as_str(r.sp.actual_sp),
            }
            for r in final_market.runners 
        ]

        # runner price data for markets that go in play
        if preplay_traded is not None:
            def runner_vals(r):
                (pre_ltp, pre_traded), (post_ltp, post_traded, sp_traded) = r

                inplay_only = list(filter(lambda ps: ps.size > 0, [
                    PriceSize(
                        price=post_ps.price, 
                        size=post_ps.size - next((pre_ps.size for pre_ps in pre_traded if pre_ps.price == post_ps.price), 0)
                    )
                    for post_ps in post_traded 
                ]))

                (ip_wavg, ip_matched, ip_min, ip_max) = parse_traded(inplay_only)
                (pre_wavg, pre_matched, pre_min, pre_max) = parse_traded(pre_traded)

                return {
                    'preplay_ltp': as_str(pre_ltp),
                    'preplay_min': as_str(pre_min),
                    'preplay_max': as_str(pre_max),
                    'preplay_wavg': as_str(pre_wavg),
                    'preplay_matched': as_str((pre_matched or 0) + (sp_traded or 0)),
                    'inplay_ltp': as_str(post_ltp),
                    'inplay_min': as_str(ip_min),
                    'inplay_max': as_str(ip_max),
                    'inplay_wavg': as_str(ip_wavg),
                    'inplay_matched': as_str(ip_matched),
                }

            runner_traded = [ runner_vals(r) for r in zip_longest(preplay_traded, postplay_traded, fillvalue=PriceSize(0, 0)) ]

        # runner price data for markets that don't go in play
        else:
            def runner_vals(r):
                (ltp, traded, sp_traded) = r
                (wavg, matched, min_price, max_price) = parse_traded(traded)

                return {
                    'preplay_ltp': as_str(ltp),
                    'preplay_min': as_str(min_price),
                    'preplay_max': as_str(max_price),
                    'preplay_wavg': as_str(wavg),
                    'preplay_matched': as_str((matched or 0) + (sp_traded or 0)),
                    'inplay_ltp': '',
                    'inplay_min': '',
                    'inplay_max': '',
                    'inplay_wavg': '',
                    'inplay_matched': '',
                }

            runner_traded = [ runner_vals(r) for r in postplay_traded ]

        # printing to csv for each runner
        for (rdata, rprices) in zip(runner_data, runner_traded):
            # defining data to go in each column
            output.write(
                "{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}\n".format(
                    postplay_market.market_id,
                    postplay_market.market_definition.market_time,
                    postplay_market.market_definition.country_code,
                    postplay_market.market_definition.venue,
                    postplay_market.market_definition.name,
                    rdata['selection_id'],
                    rdata['selection_name'],
                    rdata['selection_status'],
                    rdata['sp'],
                    rprices['preplay_min'],
                    rprices['preplay_max'],
                    rprices['preplay_wavg'],
                    rprices['preplay_ltp'],
                    rprices['preplay_matched'],
                    rprices['inplay_min'],
                    rprices['inplay_max'],
                    rprices['inplay_wavg'],
                    rprices['inplay_ltp'],
                    rprices['inplay_matched'],
                )
            )


from datetime import datetime
from typing import Iterator, Optional, Sequence, str
import betfair_data

class BflwAdapter(Iterator[BflwIter]): ...

class BflwIter(Iterator[Sequence[MarketBook]]):
    def __init__(self, path: str, bytes: bytes, cumulative_runner_tv: bool = True) -> None: ...
    file_name: str

class MarketBook:
    bet_delay: int
    bsp_reconciled: bool
    complete: bool
    cross_matching: bool
    inplay: bool
    is_market_data_delayed: bool
    last_match_time: datetime.datetime
    market_id: str
    number_of_active_runners: int
    number_of_runners: int
    number_of_winners: int
    publish_time: datetime.datetime
    publish_time_epoch: int
    runners: list[RunnerBook]
    runners_voidable: bool
    status: str
    total_available: float
    total_matched: float
    version: int
    market_definition: MarketDefinition

class RunnerBook:
    adjustment_factor: float
    ex: betfair_data.RunnerBookEX
    handicap: float
    last_price_traded: float
    removal_date: datetime.datetime
    selection_id: int
    sp: betfair_data.RunnerBookSP
    status: str
    total_matched: float
    matches: None
    orders: None


class MarketDefinition:
    bet_delay: int
    betting_type: str
    bsp_market: bool
    bsp_reconciled: bool
    complete: bool
    country_code: str
    cross_matching: bool
    discount_allowed: bool
    event_id: str
    event_type_id: str
    in_play: bool
    market_base_rate: float
    market_time: datetime.datetime
    market_type: str
    number_of_active_runners: int
    number_of_winners: int
    open_date: datetime.datetime
    persistence_enabled: bool
    regulators: str
    runners: list[MarketDefinitionRunner]
    runners_voidable: bool
    settled_time: Optional[datetime.datetime]
    status: str
    suspend_time: Optional[datetime.datetime]
    timezone: str
    turn_in_play_enabled: bool
    venue: Optional[str]
    version: int
    lineMaxUnit: Optional[float]
    lineMinUnit: Optional[float]
    lineInterval: Optional[float]
    name: Optional[str]
    eventName: Optional[str]
    priceLadderDefinition: None
    keyLineDefinition: None
    raceType: None


class MarketDefinitionRunner:
    adjustment_factor: Optional[float]
    selection_id: int
    removal_date: Optional[datetime.datetime]
    sort_priority: int
    status: str
    bsp: Optional[float]
    handicap: float
    name: Optional[str]
